#!/bin/bash
#
# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************
#

if [ "${LIB_MPI_CHECKS_SOURCED}" = "" ] ; then
LIB_MPI_CHECKS_SOURCED="true"

##############################################################################
# Check to see if mpirun is on the path and is executable.
# Globals:
#   ${VERBOSE}
# Arguments:
#   None
# Returns:
#   None
##############################################################################
mpirunTest()
{
    local mpipath=$(which mpirun 2> /dev/null)
    if ! [ -x "${mpipath}" ] ; then
        print_error "ERROR" $MSG_NO_MPIRUN
        [ "${VERBOSE}" = "true" ] && >&2 echo "Path: ${PATH}"
        exit 1
    fi
}

##############################################################################
# Checks to see if the executable 'mpitool' exists.
#  -If it does, it is used to test the mpi arguments with the 'hello world' option.
#  -Otherwise a warning is displayed, but execution continues.
# Globals:
#   ${VERBOSE}
#   ${WORKING_DIR}
# Arguments:
#   None
# Returns:
#   None
##############################################################################
mpiTest()
{
    # If mpitool exists and is executable, use that to test the specified MPI
    # options by running the mpitool "hello world" test.
    local mpiToolPath=$(which mpitool 2> /dev/null)
    if [ -x "${mpiToolPath}" ] ; then
        mpirun -timeout "${DEBUG_TIMEOUT}" "$@" mpitool -hw > "${WORKING_DIR}/mpiTest.out" 2> "${WORKING_DIR}/mpiTest.err"
        local ret_code="$?"
        if [ ${ret_code} -ne 0 ]; then
            print_error "ERROR" $MSG_MPI_OPS
            [ "${VERBOSE}" = "true" ] && >&2 cat "${WORKING_DIR}/mpiTest.err"
            exit ${ret_code}
        fi
    else
        [ "${VERBOSE}" = "true" ] && print_error "WARN" $MSG_NO_MPITOOL
    fi
}

##############################################################################
# Iterates through the list of hosts, first checking that each host is
# "'ssh'able". If it is, then attempt to ssh from that host to every other.
# Every host will be checked, even if an error is encountered.
# Globals:
#   ${VERBOSE}
#   ${WORKING_DIR}
#   ${MPI_ARGS}
# Arguments:
#   List of hosts.
# Returns:
#   None
##############################################################################
verifyHostConnectivity()
{
    local HOSTS=("$@")

    # These ssh options force the usage of private/public keys and don't prompt the user for input.
    local SSH_ARGS=("-oPasswordAuthentication=no" "-oStrictHostKeyChecking=no" "-oBatchMode=yes")

    local ENCOUNTERED_ERROR="false"
    for ((i=0; i<${#HOSTS[@]}; i++)); do
        mpirun -timeout "${DEBUG_TIMEOUT}" "${MPI_ARGS[@]}" -n 1 -H ${HOSTS[i]} true &> "${WORKING_DIR}/connectTest${HOSTS[i]}.out"
        if [ $? -ne 0 ]; then
            ssh ${SSH_ARGS[@]} ${HOSTS[i]} exit &> "${WORKING_DIR}/sshTest${HOSTS[i]}.out"
            local ret_code=$?
            if [ ${ret_code} -ne 0 ]; then
                ENCOUNTERED_ERROR="true"
                print_error "ERROR" $MSG_SSH_ERROR
                [ "${VERBOSE}" = "true" ] && >&2 cat "${WORKING_DIR}/sshTest${HOSTS[i]}.out"
                continue
            fi
        fi
    done

    if [ ${ENCOUNTERED_ERROR} = "true" ]; then
        exit 1
    fi
}

fi