# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018, 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************
from __future__ import print_function

import sys
import struct
import numpy as np
from scipy import sparse, stats
import os
from numba import cuda

if sys.version_info[0] < 3:
    import libsnapmllocal2 as libsnapmllocal
else:
    import _snap_ml.libsnapmllocal3 as libsnapmllocal

def get_gpu_count():
    return libsnapmllocal.get_gpu_count()

def read_from_snap_format(filename):
	with open(filename, "rb") as f:
	
		data_type = struct.unpack('I',f.read(4))[0]

		if(data_type == 0):
			# dense data
			print("Reading dense data from snap format")
			transpose = struct.unpack('?',f.read(1))[0]
			num_ex    = struct.unpack('I',f.read(4))[0]
			num_ft    = struct.unpack('I',f.read(4))[0]
			y         = np.frombuffer(f.read(4*num_ex),dtype=np.float32)
			X         = np.frombuffer(f.read(4*num_ex*num_ft),dtype=np.float32)
			# Allow for data-type change
			y.setflags(write=True)
			X.setflags(write=True)

			if transpose:
				X.shape = (num_ft,num_ex)
			else:
				X.shape = (num_ex,num_ft)

			return X,y

		elif(data_type == 1):
			# sparse data
			print("Reading sparse data from snap format")
			transpose = struct.unpack('?',f.read(1))[0]
			num_ex    = struct.unpack('I',f.read(4))[0]
			num_ft    = struct.unpack('I',f.read(4))[0]

			if transpose:
				count = np.frombuffer(f.read(4*num_ft),dtype=np.uint32)
			else:
				count = np.frombuffer(f.read(4*num_ex),dtype=np.uint32)

			num_nz = np.sum(count)
			indptr = np.cumsum(np.insert(count,0,0))
			num_bytes = 4*num_nz
			y      = np.frombuffer(f.read(4*num_ex), dtype=np.float32)
			ind    = np.frombuffer(f.read(num_bytes.astype(np.uint64)), dtype=np.uint32)
			val    = np.frombuffer(f.read(num_bytes.astype(np.uint64)), dtype=np.float32)
			# Allow for data-type change
			y.setflags(write=True)
			val.setflags(write=True)

			if transpose:
				X = sparse.csc.csc_matrix((val,ind,indptr),shape=(num_ex,num_ft))
			else:
				X = sparse.csr.csr_matrix((val,ind,indptr),shape=(num_ex,num_ft))

			return X,y

		elif(data_type == 2):
			# l2-sparse data
			raise Exception("Sparse data with implicit values not supported by numpy/scipy")
		else:
			# not recognized
			raise Exception("Unrecognized data type")
	
		


def dump_to_snap_format(X, y, filename, transpose=False, implicit_vals=False):

	num_ex = X.shape[0]
	num_ft = X.shape[1]
	assert y.shape[0] == num_ex

	assert type(y) is np.ndarray
	y = y.astype(np.float32)

	if(type(X) is np.ndarray):
		print("Writing Dense dataset to snap format")

		# deep copy 
		if transpose:
			X = X.T.copy()

		X = X.astype(np.float32)
		
		with open(filename, "wb") as f:
			f.write(struct.pack('I', 0))
			f.write(struct.pack('?', transpose))
			f.write(struct.pack('I', num_ex))
			f.write(struct.pack('I', num_ft))
			f.write(y.data)
			f.write(X.data)

	elif(type(X) is sparse.csr.csr_matrix):

		print("Writing Sparse dataset to snap format")
		if transpose:
			X = X.tocsc()

		ind   = X.indices.astype(np.uint32) 
		val   = X.data.astype(np.float32)
		count = np.diff(X.indptr).astype(np.uint32)

		with open(filename, "wb") as f:
			if implicit_vals:
				f.write(struct.pack('I', 2))
			else:
				f.write(struct.pack('I', 1))
			f.write(struct.pack('?', transpose))
			f.write(struct.pack('I', num_ex))
			f.write(struct.pack('I', num_ft))
			f.write(count.data)
			f.write(y.data)
			f.write(ind.data)

			if not implicit_vals:
				f.write(val.data)
	else:
		print("Unrecognized type; cannot write to snap format")


'''


num_ex = 10
num_ft = 30
import numpy as np
X = np.random.rand(num_ex,num_ft)
y = np.random.rand(num_ex)


X = sparse.random(num_ex, num_ft, 0.1, format='csr', data_rvs=stats.uniform(0,1).rvs)

print(X)
Y = X.tocsc()
print(Y.indptr)
print(Y.indices)

dump_to_snap_format(X,y,"test.snap", transpose=True)

X_chk, y_chk = read_from_snap_format("test.snap")

print(X)
'''

def _is_mpi_enabled():
    val = os.getenv('OMPI_COMM_WORLD_LOCAL_RANK')
    if val is None:
        return False
    else:
        return True

def _is_data_snapml_partition_type(X):
    if ( X.__class__.__name__ in ["SparsePartition", "DensePartition", "ConstantValueSparsePartition"] ):
        return True
    return False

#CUDA kernel to convert label
@cuda.jit
def _convert_labs(dev_arry):
    rw, cl= cuda.grid(2)
    if rw < dev_arry.shape[0] and cl < dev_arry.shape[1]: 
       dev_arry[rw, cl]=2*dev_arry[rw, cl] - 1

#Verify if the objective to be solved is compatible with the input data-type
def _verify_compat_obj_dtype(is_dual, dtype):
    if ((is_dual == False) and (dtype == "csr_matrix")):
        raise ValueError("Primal objective can only be solved with transposed data.")
    if ((is_dual == True) and (dtype == "csc_matrix")):
        raise ValueError("Dual objective can only be solved with non-transposed data.")
    return True

#Predict can only be performed with a non-transposed data
def _verify_inference_dtype(dtype):
    if (dtype == "csc_matrix"):
        raise ValueError("Inference can only be performed with a non-transposed data.")
    return True

#Checks if the data-type and parameters used for DeviceNDArray are supported
def _check_devicendarray_types(X,y,self):
    DEVICE_NDARRAY_MSG = "For DeviceNDArray as input"
    
    if(type(X).__name__ == "DeviceNDArray" or type(y).__name__ == "DeviceNDArray"):
        if ((type(X).__name__ != "DeviceNDArray") or (type(y).__name__ != "DeviceNDArray")):    
            raise TypeError(DEVICE_NDARRAY_MSG + ", X_train and y_train should be of same types")
        if (X.dtype != "float32" or y.dtype != "float32"):
            raise TypeError(DEVICE_NDARRAY_MSG + ", only float32 data type is supported")
        if(len(self.device_ids_)>1):
            raise ValueError(DEVICE_NDARRAY_MSG + ", training on multiple GPUs is not supported")
        if(len(self.device_ids_)==1 and self.device_ids_[0]!=0):
            raise ValueError(DEVICE_NDARRAY_MSG + ", training on GPU id 0 is only supported")
        if (self.use_gpu_ == False):
            raise ValueError(DEVICE_NDARRAY_MSG + ", use_gpu true is mandatory option")
        if _is_mpi_enabled():
            raise ValueError(DEVICE_NDARRAY_MSG + ", MPI support is not provided")
        if(self.return_training_history_ is not None):
            raise ValueError(DEVICE_NDARRAY_MSG + "training_history is not supported")
 
    return True
