# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************



from pai4sk.linear_model.coordinate_descent import LassoSklearn
from ...utils.validation import check_array, check_X_y
import warnings
import numbers
import numpy as np
from .utils import _check_snapml_types, _check_snapml_type, _is_gpu_matrix_type, _get_data_size, _check_uparams
from _snap_ml.utils import _is_mpi_enabled, _is_data_snapml_partition_type
_mpi_enabled = _is_mpi_enabled()


class Lasso(LassoSklearn):
    """Linear Model trained with L1 prior as regularizer (aka the Lasso)

    The optimization objective for Lasso is::

        (1 / (2 * n_samples)) * ||y - Xw||^2_2 + alpha * ||w||_1

    Technically the Lasso model is optimizing the same objective function as
    the Elastic Net with ``l1_ratio=1.0`` (no L2 penalty).

    Read more in the :ref:`User Guide <lasso>`.

    For SnapML solver this supports both local and distributed(MPI) method of execution.

    Parameters
    ----------
    alpha : float, optional
        Constant that multiplies the L1 term. Defaults to 1.0.
        ``alpha = 0`` is equivalent to an ordinary least square, solved
        by the :class:`LinearRegression` object. For numerical
        reasons, using ``alpha = 0`` with the ``Lasso`` object is not advised.
        Given this, you should use the :class:`LinearRegression` object.

    fit_intercept : boolean, optional, default True
        Whether to calculate the intercept for this model. If set
        to False, no intercept will be used in calculations
        (e.g. data is expected to be already centered).

    normalize : boolean, optional, default False
        This parameter is ignored when ``fit_intercept`` is set to False.
        If True, the regressors X will be normalized before regression by
        subtracting the mean and dividing by the l2-norm.
        If you wish to standardize, please use
        :class:`pai4sk.preprocessing.StandardScaler` before calling ``fit``
        on an estimator with ``normalize=False``.

    precompute : True | False | array-like, default=False
        Whether to use a precomputed Gram matrix to speed up
        calculations. If set to ``'auto'`` let us decide. The Gram
        matrix can also be passed as argument. For sparse input
        this option is always ``True`` to preserve sparsity.

    copy_X : boolean, optional, default True
        If ``True``, X will be copied; else, it may be overwritten.

    max_iter : int, optional
        The maximum number of iterations

    tol : float, optional
        The tolerance for the optimization: if the updates are
        smaller than ``tol``, the optimization code checks the
        dual gap for optimality and continues until it is smaller
        than ``tol``.

    warm_start : bool, optional
        When set to True, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution.
        See :term:`the Glossary <warm_start>`.

    positive : bool, optional
        When set to ``True``, forces the coefficients to be positive.

    random_state : int, RandomState instance or None, optional, default None
        The seed of the pseudo random number generator that selects a random
        feature to update.  If int, random_state is the seed used by the random
        number generator; If RandomState instance, random_state is the random
        number generator; If None, the random number generator is the
        RandomState instance used by `np.random`. Used when ``selection`` ==
        'random'.

    selection : str, default 'cyclic'
        If set to 'random', a random coefficient is updated every iteration
        rather than looping over features sequentially by default. This
        (setting to 'random') often leads to significantly faster convergence
        especially when tol is higher than 1e-4.

    verbose : bool, default : False
        If True, it prints the training cost, one per iteration. Warning: this will increase the
        training time. For performance evaluation, use verbose=False.

    use_gpu : bool, default : True
        Flag for indicating the hardware platform used for training. If True, the training
        is performed using the GPU. If False, the training is performed using the CPU.
        The value of this parameter is subjected to changed based on the training data unless set explicitly.
        Applicable only for snapml solver

    device_ids : array-like of int, default : []
        If use_gpu is True, it indicates the IDs of the GPUs used for training.
        For single GPU training, set device_ids to the GPU ID to be used for training, e.g., [0].
        For multi-GPU training, set device_ids to a list of GPU IDs to be used for training, e.g., [0, 1].\
        Applicable only for snapml solver

    num_threads : int, default : 1
        The number of threads used for running the training. The value of this parameter
        should be a multiple of 32 if the training is performed on GPU (use_gpu=True)
        (default value for GPU is 256). Applicable only for snapml solver

    return_training_history : str or None, default : None
        How much information about the training should be collected and returned by the fit function. By 
        default no information is returned (None), but this parameter can be set to "summary", to obtain 
        summary statistics at the end of training, or "full" to obtain a complete set of statistics 
        for the entire training procedure. Note, enabling either option will result in slower training.
        Applicable only for snapml solver

    privacy : bool, default : False
        Train the model using a differentially private algorithm.

    eta : float, default : 0.3
        Learning rate for the differentially private training algorithm.

    batch_size : int, default : 100
        Mini-batch size for the differentially private training algorithm.

    privacy_epsilon : float, default : 10.0
        Target privacy gaurantee. Learned model will be (privacy_epsilon, 0.01)-private.

    grad_clip : float, default: 1.0
        Gradient clipping parameter for the differentially private training algorithm

    Attributes
    ----------
    coef_ : array, shape (n_features,) | (n_targets, n_features)
        parameter vector (w in the cost function formula)

    sparse_coef_ : scipy.sparse matrix, shape (n_features, 1) | \
            (n_targets, n_features)
        ``sparse_coef_`` is a readonly property derived from ``coef_``

    intercept_ : float | array, shape (n_targets,)
        independent term in decision function.

    n_iter_ : int | array-like, shape (n_targets,)
        number of iterations run by the coordinate descent solver to reach
        the specified tolerance.

    training_history_ : dict
        It returns a dictionary with the following keys : 'epochs', 't_elap_sec', 'train_obj'.
        If 'return_training_history' is set to "summary", 'epochs' contains the total number of
        epochs performed, 't_elap_sec' contains the total time for completing all of those epochs.
        If 'return_training_history' is set to "full", 'epochs' indicates the number of epochs
        that have elapsed so far, and 't_elap_sec' contains the time to do those epochs.
        'train_obj' is the training loss.
        Applicable only for snapml solver.

    support_ : array-like
        Indices of the features that lie in the support ond contribute to the decision.

    model_sparsity_ : float
        Fraction of non-zeros in the model parameters.

    Examples
    --------
    >>> from pai4sk import linear_model
    >>> clf = linear_model.Lasso(alpha=0.1)
    >>> clf.fit([[0,0], [1, 1], [2, 2]], [0, 1, 2])
    Lasso(alpha=0.1, copy_X=True, fit_intercept=True, max_iter=1000,
       normalize=False, positive=False, precompute=False, random_state=None,
       selection='cyclic', tol=0.0001, warm_start=False)
    >>> print(clf.coef_)
    [0.85 0.  ]
    >>> print(clf.intercept_)  # doctest: +ELLIPSIS
    0.15...

    See also
    --------
    lars_path
    lasso_path
    LassoLars
    LassoCV
    LassoLarsCV
    pai4sk.decomposition.sparse_encode

    Notes
    -----
    The algorithm used to fit the model is coordinate descent.

    To avoid unnecessary memory duplication the X argument of the fit method
    should be directly passed as a Fortran-contiguous numpy array.
    """
    __slots__ = ('_lassosklearn')

    _param_list = []    # Store original arguments
    _uparams = None     # Store user provided arguments

    # Override the existing function to include the current list of parameters
    # as the init signature is modified by the decorator
    def _get_param_names(self):
        return self._param_list

    @_check_uparams
    def __init__(self, alpha=1.0, fit_intercept=True, normalize=False,
                 precompute=False, copy_X=True, max_iter=1000,
                 tol=1e-4, warm_start=False, positive=False,
                 random_state=None, selection='cyclic',
                 verbose=0,                         # pai4sk
                 use_gpu=True,                      # pai4sk
                 device_ids=[],                     # pai4sk
                 return_training_history=None,      # pai4sk
                 privacy=False,                     # pai4sk
                 eta=0.3,                           # pai4sk
                 batch_size=100,                    # pai4sk 
                 privacy_epsilon=10,                # pai4sk
                 grad_clip=1,                       # pai4sk
                 num_threads=1):                    # pai4sk
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.normalize = normalize
        self.precompute = precompute
        self.max_iter = max_iter
        self.copy_X = copy_X
        self.tol = tol
        self.warm_start = warm_start
        self.positive = positive
        self.random_state = random_state
        self.selection = selection  
        self.use_gpu = use_gpu
        self.device_ids = device_ids
        self.num_threads = num_threads
        self.verbose = verbose
        self.return_training_history = return_training_history
        self.privacy = privacy
        self.eta = eta
        self.batch_size = batch_size
        self.privacy_epsilon = privacy_epsilon
        self.grad_clip = grad_clip
        super(Lasso, self).__init__(
            alpha=alpha, fit_intercept=fit_intercept,
            normalize=normalize, precompute=precompute, copy_X=copy_X,
            max_iter=max_iter, tol=tol, warm_start=warm_start,
            positive=positive, random_state=random_state,
            selection=selection)
        self._lassosklearn = super(Lasso, self)
        self._snaplr = None               
        
    def __setstate__(self, state):
        self.__dict__.update(state)
        self._lassosklearn = super(Lasso, self).__init__(
            alpha=alpha, fit_intercept=fit_intercept,
            normalize=normalize, precompute=precompute, copy_X=copy_X,
            max_iter=max_iter, tol=tol, warm_start=warm_start,
            positive=positive, random_state=random_state,
            selection=selection)

    def _check_set_snapml(self, X, y):
        snap_args = {}
        mod_params = []
        unsupported_params = []

        if ( not self.random_state is None):
            unsupported_params.append("random_state")

        if not _check_snapml_types(X,y):
            unsupported_params.append("X/y")
        else:
            # choose CPU implementation for smaller dataset (in-memory size < 400kb)
            if not _is_data_snapml_partition_type(X) and not _is_gpu_matrix_type(X, y):
                if _get_data_size(X) < 400 and not 'use_gpu' in self._uparams :
                    self.use_gpu = False
                    mod_params.append("use_gpu")

                # can't handle multi-variate
                if not y.ndim == 1:
                    unsupported_params.append("y")

        if not unsupported_params:
            # Check for GPUs
            snap_args['use_gpu'] = False if self.use_gpu is None else self.use_gpu
            snap_args['device_ids'] = [] if self.device_ids is None else self.device_ids
            if self.use_gpu:
                from pai4sk import sml_utils
                ndevices = sml_utils.get_gpu_count()
                if ndevices == 0:
                    if 'use_gpu' in self._uparams:
                        raise ValueError("SnapML: GPU device not found")
                    mod_params.append("use_gpu")
                    self.use_gpu = False

            if 'max_iter' in self._uparams:
                snap_args['max_iter'] = self.max_iter
            else:
                mod_params.append("max_iter")
                snap_args['max_iter'] = 1000

            if 'tol' in self._uparams:
                snap_args['tol'] = self.tol
            else:
                mod_params.append("tol")
                snap_args['tol'] = 0.001

            if 'fit_intercept' in self._uparams:
                snap_args['fit_intercept'] = self.fit_intercept
            else:
                mod_params.append("fit_intercept")
                self.fit_intercept = False

            snap_args['verbose'] = False if self.verbose == 0 else True
            snap_args['regularizer'] = 1 if self.alpha is None else self.alpha
            snap_args['intercept_scaling'] = 1.0 if self.normalize is None else self.normalize
            snap_args['return_training_history'] = self.return_training_history
            snap_args['num_threads'] = self.num_threads
            snap_args['dual'] = False
            snap_args['penalty'] = "l1"
            snap_args['privacy'] = self.privacy
            snap_args['eta'] = self.eta
            snap_args['batch_size'] = self.batch_size
            snap_args['privacy_epsilon'] = self.privacy_epsilon
            snap_args['grad_clip'] = self.grad_clip

            from pai4sk import LinearRegression as _snaplr
            self._snaplr = _snaplr(**snap_args)

            warnings.warn("SnapML: Default solver is set to 'snapml'.")

        if unsupported_params:
            warnMsg = "SnapML: fallback to sklearn implementation due to unsupported parameters/values: " + ", ".join([str(x) for x in unsupported_params])
            warnings.warn(warnMsg, UserWarning)
            return False

        if mod_params:
            warnMsg = "SnapML: Default values for these parameters are modified for 'snapml' solver: " + ", ".join([str(x) for x in mod_params])
            warnings.warn(warnMsg, UserWarning)

        return True

    def fit(self, X, y, check_input=True):
        """Fit model with coordinate descent.

        Parameters
        -----------
        X : ndarray or scipy.sparse matrix, (n_samples, n_features)
            Data
            For SnapML solver it also supports input of types SnapML data partition and DeviceNDArray.

        y : ndarray, shape (n_samples,) or (n_samples, n_targets)
            Target. Will be cast to X's dtype if necessary

        check_input : boolean, (default=True)
            Allow to bypass several input checking.
            Don't use this parameter unless you know what you do.

        Notes
        -----

        Coordinate descent is an algorithm that considers each column of
        data at a time hence it will automatically convert the X input
        as a Fortran-contiguous numpy array if necessary.

        To avoid memory re-allocation it is advised to allocate the
        initial data in memory directly using that format.
        """

        if self._check_set_snapml(X, y):
            self.training_history_ = self._snaplr.fit(X, y)
            
            if not _mpi_enabled:
                # Attributes only supported with snap-ml-local
                self.n_iter_ = self._snaplr.n_iter_
                self.support_ = self._snaplr.support_
                self.model_sparsity_ = self._snaplr.model_sparsity_

            if (self.fit_intercept):
                self.coef_ = self._snaplr.model_[:,:-1]
                self.intercept_ = self._snaplr.model_[:,-1]
            else:
                self.coef_ = self._snaplr.model_
                self.intercept_ = np.empty(shape=())

            if y.ndim == 1:
                self.coef_ = np.ravel(self.coef_)

            return self
        else:
            ls_sk = self._lassosklearn.fit(X, y, check_input)
            self.coef_ = ls_sk.coef_
            self.intercept_ = ls_sk.intercept_
            self.n_iter_ = ls_sk.n_iter_
            self.training_history_ = []
            return self


    def predict(self, X, num_threads=0):

        """
        Class predictions
        The returned class estimates.
        Parameters
        ----------
        X : sparse matrix (csr_matrix) or dense matrix (ndarray)
            Dataset used for predicting class estimates.
            For SnapML solver it also supports input of type SnapML data partition.
       num_threads : int, default : 0
            Number of threads used to run inference.
            By default inference runs with maximum number of available threads.
        Returns
        -------
        proba: array-like, shape = (n_samples,)
            Returns the predicted class of the sample.
        """
        if not hasattr(self, "coef_"):
            raise NotFittedError("Call fit before prediction")
        
        if _check_snapml_type(X) and self._snaplr: 
            return self._snaplr.predict(X, num_threads)
        else:
            return self._lassosklearn.predict(X)
